// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright 2020 RnD Center "ELVEES", JSC
 */

#include <linux/debugfs.h>
#include <linux/uaccess.h>

#include "elcore50-core.h"
#include "elcore50-debug.h"
#include "elcore50-pm.h"
#include "elcore50-proc.h"
#include "elcore50-syscall.h"

static u32 uptime_timer_period_msec = 500;

static enum hrtimer_restart timer_callback(struct hrtimer *timer)
{
	struct elcore50_core *core = from_timer(core, timer, timer);
	struct cluster_priv *drv_priv = core->drv_priv;
	unsigned long clk_rate = drv_priv->devfreq_clk_last_value;
	unsigned long flags;
	u64 total_ticks, cur_ticks = 0;

	hrtimer_forward_now(timer, ms_to_ktime(uptime_timer_period_msec));

	mutex_lock(&drv_priv->rpm_lock);
	if (pm_runtime_active(drv_priv->dev))
		cur_ticks = elcore50_read64(core, DSP_TIC_CNTR);
	mutex_unlock(&drv_priv->rpm_lock);

	spin_lock_irqsave(&core->ticks_lock, flags);
	total_ticks = core->ticks_acc + (cur_ticks - core->last_ticks);
	core->ticks_acc = 0;
	core->last_ticks = cur_ticks;
	spin_unlock_irqrestore(&core->ticks_lock, flags);

	// percentage * 100
	core->current_usage = (total_ticks * 1000 * 100 * 100) /
			      (uptime_timer_period_msec * clk_rate);

	return HRTIMER_RESTART;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 253)
int dev_err_probe(const struct device *dev, int err, const char *fmt, ...)
{
	struct va_format vaf;
	va_list args;

	va_start(args, fmt);
	vaf.fmt = fmt;
	vaf.va = &args;

	if (err != -EPROBE_DEFER)
		dev_err(dev, "error %pe: %pV", ERR_PTR(err), &vaf);
	else
		dev_dbg(dev, "error %pe: %pV", ERR_PTR(err), &vaf);

	va_end(args);

	return err;
}
#endif

void elcore50_get_platform_info(struct elcore50_core *core,
				struct elcore50_device_info *device_info)
{
	device_info->nclusters = IDR_NCLUSTERS(core->idr);
	device_info->cluster_id = IDR_CLUSTER(core->idr);
	device_info->cluster_cap = IDR_CLUSTERCAP(core->idr);
	device_info->core_in_cluster_id = IDR_CORENUM(core->idr);
}

int elcore50_get_core_idx(struct elcore50_core *core, void __user *arg)
{
	struct elcore50_device_info device_info;
	int ret;

	if (core == NULL) {
		// Possible global job instance is still not scheduled on DSP
		return -EPERM;
	}

	elcore50_get_platform_info(core, &device_info);

	ret = copy_to_user(arg, &device_info,
			   sizeof(struct elcore50_device_info));
	if (ret)
		return ret;

	return 0;
}

int elcore50_get_caps(struct elcore50_core *core, void __user *arg)
{
	struct elcore_caps elcore_caps;
	int ret;

	strcpy(elcore_caps.drvname, "elcore50");
	elcore_caps.hw_id = core->idr;
	elcore_caps.hw_id &= (IDR_VER_MASK | IDR_REV_MASK);
	elcore_caps.hw_id = elcore_caps.hw_id >> IDR_REV_OFFT;

	ret = copy_to_user(arg, &elcore_caps, sizeof(struct elcore_caps));
	if (ret)
		return ret;

	return 0;
}

static int elcore50_open(struct inode *inode, struct file *file)
{
	struct elcore50_core *core;
	int ret;

	core = container_of(inode->i_cdev, struct elcore50_core, cdev);

	file->private_data = core;

	ret = pm_runtime_get_sync(core->drv_priv->dev);

	return ret < 0 ? ret : 0;
}

static long elcore50_ioctl(struct file *file, unsigned int cmd,
			   unsigned long arg)
{
	struct elcore50_core *pdata =
		(struct elcore50_core *)file->private_data;
	void __user *const uptr = (void __user *)arg;

	switch (cmd) {
	case ELCORE50_IOC_CREATE_JOB:
		return elcore50_create_job(file, pdata->drv_priv->dev, uptr);
	case ELCORE50_IOC_ENQUEUE_JOB:
		return elcore50_enqueue_job_inst(pdata, uptr);
	case ELCORE50_IOC_GET_JOB_STATUS:
		return elcore50_get_job_inst_status(uptr);
	case ELCORE50_IOC_GET_JOB_COUNT:
		return elcore50_get_job_inst_count(pdata, uptr);
	case ELCORE50_IOC_GET_JOB_LIST:
		return elcore50_get_job_inst_list(pdata, uptr);
	case ELCORE50_IOC_DBG_JOB_ATTACH:
		return elcore50_job_dbg_attach(pdata, uptr);
	case ELCORE50_IOC_GET_CORE_IDX:
		return elcore50_get_core_idx(pdata, uptr);
	case ELCORE50_IOC_CREATE_MAPPER:
		/* Deprecated IOCTL */
		return elcore50_create_mapper(pdata->drv_priv->dev, uptr);
	case ELCORE50_IOC_CREATE_BUFFER:
		/* Deprecated IOCTL */
		return elcore50_create_buffer(pdata->drv_priv->dev, uptr);
	case ELCORE50_IOC_SYNC_BUFFER:
		/* Deprecated IOCTL */
		return elcore50_sync_buffer(file, uptr);
	case ELCIOC_GET_CAPS:
		return elcore50_get_caps(pdata, uptr);
	}
	return -ENOTTY;
}

static int elcore50_release(struct inode *inode, struct file *file)
{
	struct elcore50_core *core;

	core = container_of(inode->i_cdev, struct elcore50_core, cdev);

	mutex_lock(&core->drv_priv->rpm_lock);
	pm_runtime_put(core->drv_priv->dev);
	mutex_unlock(&core->drv_priv->rpm_lock);

	return 0;
}

static const struct file_operations elcore50_fops = {
	.owner = THIS_MODULE,
	.open = elcore50_open,
	.unlocked_ioctl = elcore50_ioctl,
	.release = elcore50_release,
};

void elcore50_core_abort(struct elcore50_core *core)
{
	/* switch off device interrupts and force-stop it */
#ifndef ELCORE50_NO_IRQS
	elcore50_write(0, core, DSP_DMASKR);
#endif
	elcore50_write(DCSR_STOP, core, DSP_DCSR);
}

void print_dump(struct elcore50_core *core)
{
	u32 reg_addr, i, j;
	u64 tmp[32];

	if (reg_dump_enable == 0)
		return;

	/* RF */
	reg_addr = DSP_R0;
	for (i = 0; i < 32; i += 4) {
		dev_warn(core->dev,
			 "RF%02d:      %016llx %016llx %016llx %016llx\n", i,
			 elcore50_read64(core, reg_addr + 8 * i),
			 elcore50_read64(core, reg_addr + 8 * (i + 1)),
			 elcore50_read64(core, reg_addr + 8 * (i + 2)),
			 elcore50_read64(core, reg_addr + 8 * (i + 3)));
	}

	/* PCU0 */
	dev_warn(core->dev, "DCSR:      %08x IDR:     %08x\n",
		 elcore50_read(core, DSP_DCSR), elcore50_read(core, DSP_IDR));

	/* PCU2 */
	dev_warn(core->dev,
		 "PC:        %08x IVAR:   %08x IRTAR:   %08x EVAR: %08x\n",
		 elcore50_read(core, DSP_PC), elcore50_read(core, DSP_IVAR),
		 elcore50_read(core, DSP_IRTAR), elcore50_read(core, DSP_EVAR));
	dev_warn(core->dev, "ERTAR:     %08x DVAR:    %08x DRTAR:  %08x\n",
		 elcore50_read(core, DSP_ERTAR), elcore50_read(core, DSP_DVAR),
		 elcore50_read(core, DSP_DRTAR));

	/* PCU1 */
	dev_warn(core->dev, "TMR:       %016llx TMRC:  %016llx\n",
		 elcore50_read64(core, DSP_TMR),
		 elcore50_read64(core, DSP_TMRC));
	dev_warn(core->dev,
		 "TCNTR:     %016llx ICNTR: %016llx CCNTR: %016llx\n",
		 elcore50_read64(core, DSP_TIC_CNTR),
		 elcore50_read64(core, DSP_INSTR_CNTR),
		 elcore50_read64(core, DSP_CMD_CNTR));

	/* PCU3 */
	reg_addr = DSP_P0;
	dev_warn(core->dev, "P0:        %08x %08x %08x %08x\n",
		 elcore50_read(core, reg_addr),
		 elcore50_read(core, reg_addr + 4),
		 elcore50_read(core, reg_addr + 8),
		 elcore50_read(core, reg_addr + 12));
	dev_warn(core->dev, "P4:        %08x %08x %08x %08x\n",
		 elcore50_read(core, reg_addr + 16),
		 elcore50_read(core, reg_addr + 20),
		 elcore50_read(core, reg_addr + 24),
		 elcore50_read(core, reg_addr + 28));

	reg_addr = DSP_VP0;
	dev_warn(core->dev, "VP0:       %016llx %016llx %016llx %016llx\n",
		 elcore50_read64(core, reg_addr),
		 elcore50_read64(core, reg_addr + 8),
		 elcore50_read64(core, reg_addr + 16),
		 elcore50_read64(core, reg_addr + 24));
	dev_warn(core->dev, "VP4:       %016llx %016llx %016llx %016llx\n",
		 elcore50_read64(core, reg_addr + 32),
		 elcore50_read64(core, reg_addr + 40),
		 elcore50_read64(core, reg_addr + 48),
		 elcore50_read64(core, reg_addr + 56));

	/* PCU4 */
	dev_warn(core->dev, "CSP:       %08x\n", elcore50_read(core, DSP_CSP));
	dev_warn(core->dev, "LC0:       %08x %08x %08x %08x\n",
		 elcore50_read(core, DSP_LC0), elcore50_read(core, DSP_LC1),
		 elcore50_read(core, DSP_LC2), elcore50_read(core, DSP_LC3));
	dev_warn(core->dev, "LB0:       %08x %08x %08x %08x\n",
		 elcore50_read(core, DSP_LB0), elcore50_read(core, DSP_LB1),
		 elcore50_read(core, DSP_LB2), elcore50_read(core, DSP_LB3));
	dev_warn(core->dev, "LA0:       %08x %08x %08x %08x\n",
		 elcore50_read(core, DSP_LA0), elcore50_read(core, DSP_LA1),
		 elcore50_read(core, DSP_LA2), elcore50_read(core, DSP_LA3));

	/* PCU6 */
	dev_warn(core->dev,
		 "IRQ_INDEX: %08x IRQR:  %08x IMASKR: %08x DQSTR: %08x\n",
		 elcore50_read(core, DSP_IRQ_INDEX),
		 elcore50_read(core, DSP_IRQR), elcore50_read(core, DSP_IMASKR),
		 elcore50_read(core, DSP_DQSTR));
	dev_warn(core->dev, "DMASKR:    %08x INDEX: %08x UIPC:   %08x\n",
		 elcore50_read(core, DSP_DMASKR),
		 elcore50_read(core, DBG_INDEX), elcore50_read(core, DSP_UIPC));
	dev_warn(core->dev, "BAPC:      %016llx\n",
		 elcore50_read64(core, DSP_BAPC));

	/* PCU10 */
	dev_warn(core->dev, "M0:        %08x %08x\n",
		 elcore50_read(core, DSP_M0), elcore50_read(core, DSP_M1));

	/* PCU12 */
	dev_warn(core->dev, "CTRL:      %08x INVCTRL:  %08x INVADDR:  %08x\n",
		 elcore50_read(core, DSP_CTRL),
		 elcore50_read(core, DSP_INVCTRL),
		 elcore50_read(core, DSP_INVADDR));
	dev_warn(core->dev, "ASID:      %08x CREGIONS: %08x MREGIONS: %08x\n",
		 elcore50_read(core, DSP_ASID),
		 elcore50_read(core, DSP_CREGIONS),
		 elcore50_read(core, DSP_MREGIONS));
	dev_warn(
		core->dev,
		"SWAITREG0: %08x SWAITREG1: %08x VWAITREG0: %08x VWAITREG1: %08x\n",
		elcore50_read(core, DSP_SWAITREG0),
		elcore50_read(core, DSP_SWAITREG1),
		elcore50_read(core, DSP_VWAITREG0),
		elcore50_read(core, DSP_VWAITREG1));

	/* Caches */
	dev_warn(
		core->dev,
		"L0DC_CTRL: %08x L1IC_CTRL: %08x L1DC_CTRL: %08x L2_CTRL:   %08x\n",
		elcore50_read(core, L0DC_CTRL), elcore50_read(core, L1IC_CTRL),
		elcore50_read(core, L1DC_CTRL), elcore50_read(core, L2_CTRL));

	/* VF */
	for (i = 0, reg_addr = DSP_VF_START; i < 64; ++i, reg_addr += 0x40) {
		for (j = 0; j < 8; ++j)
			tmp[j] = elcore50_read64(core, reg_addr + j * 8);
		dev_warn(core->dev,
			 "VF%02dH:     %016llx %016llx %016llx %016llx\n", i,
			 tmp[7], tmp[6], tmp[5], tmp[4]);
		dev_warn(core->dev,
			 "VF%02dL:     %016llx %016llx %016llx %016llx\n", i,
			 tmp[3], tmp[2], tmp[1], tmp[0]);
	}

	/* VMMU */
	dev_warn(core->dev, "PTW_PBA_L: %08x PTW_PBA_H: %08x\n",
		 elcore50_read(core, VMMU_PTW_PBA_L),
		 elcore50_read(core, VMMU_PTW_PBA_H));
	dev_warn(core->dev, "PTW_CFG:   %08x TLB_CTRL:  %08x\n",
		 elcore50_read(core, VMMU_PTW_CFG),
		 elcore50_read(core, VMMU_TLB_CTRL));

	dev_warn(core->dev, "TLBXCPT:   NUM ADDR     TYPE\n");
	for (i = 0; i < VMMU_TLBS; ++i) {
		elcore50_write(i, core, VMMU_TLBXCPT_NUM);
		dev_warn(core->dev, "           %d   %08x %d\n",
			 elcore50_read(core, VMMU_TLBXCPT_NUM),
			 elcore50_read(core, VMMU_TLBXCPT_ADDR),
			 elcore50_read(core, VMMU_TLBXCPT_TYPE));
	}
}

void elcore50_core_reset(struct elcore50_core *core)
{
	do {
		elcore50_write(DCSR_RESET, core, DSP_DCSR);
	} while (elcore50_read(core, DSP_MBARREG) != 0);
}

static int elcore50_core_init(struct elcore50_core *core,
			      struct cluster_priv *drv_priv, unsigned int major,
			      unsigned int minor, int corenum)
{
	struct device *dev;
	u32 regval;
	int core_system_idx, ret;

#ifdef ELCORE50_MCOM03
	core->xyram = drv_priv->mem;
	core->dsp_regs = drv_priv->regs;
#else
	core->xyram = drv_priv->mem + SZ_2M * corenum;
	core->dsp_regs = drv_priv->regs + DSPNEXT_OFF * corenum;
#endif
	core->pram = core->xyram + SZ_1M;

	regval = ioread32(core->dsp_regs + DSP_IDR);
	core_system_idx = IDR_CLUSTER(regval) * IDR_CLUSTERCAP(regval) +
			  IDR_CORENUM(regval);
	core->idr = regval;
	core->system_idx = core_system_idx;

	WARN_ON(scheduler_data.core[core_system_idx] != NULL);

	scheduler_data.core[core_system_idx] = core;

	core->drv_priv = drv_priv;
	cdev_init(&core->cdev, &elcore50_fops);
	ret = cdev_add(&core->cdev, MKDEV(major, minor), 1);
	if (ret < 0) {
		dev_err_probe(drv_priv->dev, ret,
			      "Failed to add ELcore-50 cdev\n");
		goto err_none;
	}

	dev = device_create(elcore50_class, drv_priv->dev, MKDEV(major, minor),
			    NULL, "elcore%d", core_system_idx);
	if (IS_ERR(dev)) {
		/* this way we can be assured cores[i] is deallocated */
		ret = PTR_ERR(dev);
		dev_err_probe(drv_priv->dev, ret,
			      "Failed to create ELcore-50 device\n");
		goto err_cdev;
	}
	core->dev = dev;
	core->corenum = corenum;

	INIT_LIST_HEAD(&core->job_queue);
	spin_lock_init(&core->queue_lock);

	core->kthread_scheduler = kthread_run(
		kthread_handler, core, "kthread-%s", core->dev->init_name);
	if (IS_ERR(core->kthread_scheduler)) {
		ret = PTR_ERR(core->kthread_scheduler);
		goto err_device;
	}

	spin_lock_init(&core->ticks_lock);

	core->ticks_acc = 0;
	core->last_ticks = 0;
	core->current_usage = 0;

	hrtimer_init(&core->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
	core->timer.function = &timer_callback;
	hrtimer_start(&core->timer, ms_to_ktime(uptime_timer_period_msec),
		      HRTIMER_MODE_REL);

	return 0;
err_device:
	device_destroy(elcore50_class, MKDEV(major, minor));
err_cdev:
	cdev_del(&core->cdev);
err_none:
	scheduler_data.core[core_system_idx] = NULL;
	return ret;
}

static void elcore50_core_destroy(struct elcore50_core *core,
				  struct cluster_priv *drv_priv)
{
	struct elcore50_job_inst_desc *cursor, *store;

	scheduler_data.core[core->system_idx] = NULL;

	hrtimer_cancel(&core->timer);

	kthread_stop(core->kthread_scheduler);
	wake_up(&scheduler_data.scheduler_waitq);

	list_for_each_entry_safe(cursor, store, &core->job_queue, queue_node) {
		elcore50_cancel_job_inst(cursor);
	}
	device_destroy(elcore50_class, core->cdev.dev);
	cdev_del(&core->cdev);
}

int elcore50_cores_init(struct cluster_priv *drv_priv, int ncores)
{
	int ret, major, minor, i;

	ret = alloc_chrdev_region(&drv_priv->dev_num, 0, ncores, "elcore50");
	if (ret < 0)
		return dev_err_probe(drv_priv->dev, ret,
				     "Failed to allocate chrdev region\n");

	drv_priv->ncores = ncores;
	drv_priv->cores = devm_kcalloc(drv_priv->dev, ncores,
				       sizeof(struct elcore50_core),
				       GFP_KERNEL);

	major = MAJOR(drv_priv->dev_num);
	for (i = 0; i < ncores; i++) {
		minor = MINOR(drv_priv->dev_num) + i;
		ret = elcore50_core_init(&drv_priv->cores[i], drv_priv, major,
					 minor, i);
		if (ret) {
			dev_err_probe(drv_priv->dev, ret,
				      "Failed to initialize core %d\n", i);
			goto err_dev;
		}
	}
	return 0;
err_dev:
	while (i > 0) {
		i--;
		elcore50_core_destroy(&drv_priv->cores[i], drv_priv);
	}
	unregister_chrdev_region(drv_priv->dev_num, ncores);
	return ret;
}

void elcore50_cores_destroy(struct cluster_priv *drv_priv)
{
	int i;

	for (i = 0; i < drv_priv->ncores; i++)
		elcore50_core_destroy(&drv_priv->cores[i], drv_priv);
	unregister_chrdev_region(drv_priv->dev_num, drv_priv->ncores);
}

#ifndef ELCORE50_NO_IRQS
irqreturn_t elcore50_irq(int irq, void *priv)
{
	struct elcore50_core *core = (struct elcore50_core *)priv;
	struct elcore50_job_inst_desc *desc;
	unsigned long flags;
	int found = 0;

	spin_lock_irqsave(&scheduler_data.running_list_lock, flags);
	list_for_each_entry(desc, &scheduler_data.running_job_list,
			    queue_node) {
		if (desc->core == core) {
			found = 1;
			break;
		}
	}
	spin_unlock_irqrestore(&scheduler_data.running_list_lock, flags);

	if (!found) {
		dev_err(core->dev, "IRQ received, but no jobs in queue!\n");
		return IRQ_NONE;
	}

	elcore50_write(0, core, DSP_DMASKR);

	desc->core_stopped_by_irq = 1;
	desc->irq_processed = 0;

	if (job_status_dump_enable)
		dev_info(core->dev, "IRQ");

	wake_up(&desc->irq_waitq);
	return IRQ_HANDLED;
}
#endif

/**
 * elcore50_list_count - count nodes in the list
 * @head:	the head for your list.
 */
int elcore50_list_count(struct list_head *head)
{
	struct list_head *pos;
	size_t count = 0;

	list_for_each(pos, head)
		count++;

	return count;
}
