// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright 2020 RnD Center "ELVEES", JSC
 */

#include <linux/slab.h>

#include "elcore50-job.h"
#include "elcore50-mmu.h"
#include "elcore50-trace.h"

struct e50_protected_region {
	u32 base_addr;
	bool is_cluster_specific;
	u32 size;
	struct userptr_mapper mapper;
};

#ifdef ELCORE50_MCOM03
static struct e50_protected_region protected_regions[] = {
	{ 0x1800000, 0, 0x800000, { 0 } }, // EPort
	{ 0x2800000, 0, 0x40000, { 0 } }, // Registers
	{ 0x2E00000, 0, 0x600000, { 0 } } // PRAM/XYRAM
};
#else
static struct e50_protected_region protected_regions[] = {
	{ 0x10000000, 0, 0x2000000, { 0 } }, // Registers
	{ 0x13000000, 1, 0x100000, { 0 } }, // QPRAM/QYRAM
	{ 0x17000000, 1, 0x40000, { 0 } } // QRegisters
};
#endif

#define JOB_DESC_TO_DEV(job_desc) (job_desc->dev)

static inline unsigned long elcore50_mmu_alloc_fixed(struct mmu_pool *pool,
						     size_t size,
						     unsigned long address)
{
	struct mmupool_data_fixed data;

	data.offset = address - ELCORE50_VADDR_START;
	return mmu_pool_alloc_algo(pool, size, mmu_pool_fixed_alloc, &data);
}

static inline void write_pte_entry(uint64_t *pte, dma_addr_t paddr,
				   enum pte_type type)
{
	uint64_t tmp;

	tmp = E50_PTE_V_MASK;
	tmp |= (paddr >> E50_PAGE_SHIFT) << E50_PTE_PADDR_SHIFT;
	tmp |= (type << E50_PTE_TYPE_SHIFT) & E50_PTE_TYPE_MASK;
	/* FIXME: this is a hack - rf#10482 */
	if (type > PTE_NEXTGL)
		tmp |= E50_PTE_D_MASK | E50_PTE_R_MASK;
	*pte = tmp;
	pr_debug("pte %p paddr %llx type=%i res=%llx", pte, paddr, type, tmp);
}

static struct page_entry *
get_page_entry(struct device *dev, struct page_entry *p_top,
	       struct list_head *pentry_list, uint64_t vaddr_mmu_dsp,
	       enum elcore50_page_size ps, int is_argument, bool is_48bit)
{
	int i, j;
	dma_addr_t paddr;
	struct page *page;
	struct page_entry *p = p_top, *p_cur;
	uint64_t *pte;
	uint64_t index;
	/* 48-bit addresses don't work with elcore50 */
	if (vaddr_mmu_dsp >> (is_48bit ? 48 : 32)) {
		dev_err(dev, "Virtual address 0x%llx out of bounds\n",
			vaddr_mmu_dsp);
		return NULL;
	}
	for (j = VMMU_PAGE_MAX_LEVEL; j > ps; j--) {
		index = (vaddr_mmu_dsp >> (E50_PAGE_SHIFT + j * 9)) &
			GENMASK(8, 0);

		p_cur = p + index;
		if (p_cur->next_lvl) {
			p = p_cur->next_lvl;
		} else {
			p_cur->next_lvl = kcalloc(E50_PTE_ENTRIES,
						  sizeof(struct page_entry),
						  GFP_KERNEL);
			if (!p_cur->next_lvl)
				return NULL;
			p_cur->next_lvl->prev_lvl = p_cur;
			p_cur->base_pentry = p;

			/* Mark the new page table dirty */
			p_cur->dirty = 1;

			page = alloc_pages(GFP_KERNEL | __GFP_ZERO,
					   get_order(E50_PAGE_SIZE));
			if (!page)
				goto clean_next_lvl;
			pte = page_address(page);
			paddr = dma_map_single_attrs(dev, pte, E50_PAGE_SIZE,
						     DMA_TO_DEVICE,
						     DMA_ATTR_SKIP_CPU_SYNC);
			if (!paddr)
				goto clean_page;
			p_cur->dma_addr = paddr;
			write_pte_entry(p_cur->pte, paddr, PTE_NEXT);
			/* Mark current page table dirty */
			if (p->prev_lvl)
				p->prev_lvl->dirty = 1;

			p->num_non_zero_ptes += 1;
			p = p_cur->next_lvl;
			p_cur->argument = is_argument;
			list_add_tail(&p_cur->list_node, pentry_list);

			for (i = 0; i < E50_PTE_ENTRIES; i++)
				p[i].pte = &pte[i];
			continue;
clean_page:
			free_pages((unsigned long)pte,
				   get_order(E50_PAGE_SIZE));
clean_next_lvl:
			kfree(p_cur->next_lvl);
			p_cur->next_lvl = NULL;
			return NULL;
		}
	}
	/* Increment ptes counter for leaf node. write_pte_entry() will be
	 * called in write_buffer_pages()
	 */
	p->num_non_zero_ptes += 1;
	/* move the pointer to the leaf node rather than leaf page start */
	index = (vaddr_mmu_dsp >> (E50_PAGE_SHIFT + j * 9)) & GENMASK(8, 0);
	p_cur = p + index;
	p_cur->base_pentry = p;
	p_cur->argument = is_argument;
	p->prev_lvl->dirty = 1;
	return p_cur;
}

static int write_buffer_pages(struct device *dev, int hugepages,
			      struct list_head *pentry_list,
			      struct page_entry *p_top,
			      uint64_t buf_vaddr_mmu_dsp, enum pte_type type,
			      dma_addr_t buf_paddr, size_t buf_size,
			      bool is_48bit, struct userptr_mapper *mapper,
			      uint64_t vaddr_start_mmu_dsp)
{
	size_t page_size;
	enum elcore50_page_size ps;
	struct page_entry *p;

	pr_debug("buf_paddr=%llx, vaddr=%llx, size=%lx\n", buf_paddr,
		 buf_vaddr_mmu_dsp, buf_size);
	WARN_ON(buf_paddr & E50_INPAGE_MASK);
	WARN_ON(buf_vaddr_mmu_dsp & E50_INPAGE_MASK);
	WARN_ON(buf_size & E50_INPAGE_MASK);
	WARN_ON(mapper == NULL);
	while (buf_size > 0) {
		if (hugepages) {
			if (!(buf_vaddr_mmu_dsp & (SZ_1G - 1)) &&
			    buf_size >= SZ_1G) {
				page_size = SZ_1G;
				ps = ELCORE50_PAGE_1G;
			} else if (!(buf_vaddr_mmu_dsp & (SZ_2M - 1)) &&
				   buf_size >= SZ_2M) {
				page_size = SZ_2M;
				ps = ELCORE50_PAGE_2M;
			} else {
				page_size = SZ_4K;
				ps = ELCORE50_PAGE_4K;
			}
		} else {
			page_size = SZ_4K;
			ps = ELCORE50_PAGE_4K;
		}
		p = get_page_entry(dev, p_top, pentry_list, buf_vaddr_mmu_dsp,
				   ps, mapper->is_argument, is_48bit);
		if (!p)
			return -ENOMEM;
		p->mapper = mapper;
		p->offset = buf_vaddr_mmu_dsp - vaddr_start_mmu_dsp;
		list_add_tail(&p->list_node, pentry_list);

		write_pte_entry(p->pte, buf_paddr, type);
		buf_vaddr_mmu_dsp += page_size;
		buf_paddr += page_size;
		buf_size -= page_size;
	}
	return 0;
}

void mmu_free_page_entries(struct device *dev, struct list_head *pentry_list,
			   int arguments_only)
{
	struct page_entry *p_cur, *store, *p;

	list_for_each_entry_safe_reverse(p_cur, store, pentry_list, list_node) {
		if (arguments_only && !p_cur->argument)
			break;

		*p_cur->pte = 0;

		p = p_cur->base_pentry;
		if (p->prev_lvl)
			p->prev_lvl->dirty = 1;
		p->num_non_zero_ptes -= 1;

		list_del(&p_cur->list_node);

		if (p_cur->dma_addr) {
			dma_unmap_single_attrs(dev, p_cur->dma_addr,
					       E50_PAGE_SIZE, DMA_FROM_DEVICE,
					       DMA_ATTR_SKIP_CPU_SYNC);
		}

		if (p->num_non_zero_ptes == 0) {
			free_pages((unsigned long)p->pte,
				   get_order(E50_PAGE_SIZE));
			if (p->prev_lvl) {
				p->prev_lvl->next_lvl = NULL;
				p->prev_lvl->mapper = NULL;
			}
			kfree(p);
		}
	}
}

void sync_elcore_pte_table_for_device(struct device *dev,
				      struct list_head *list)
{
	struct page_entry *p_cur;

	list_for_each_entry(p_cur, list, list_node) {
		if (!p_cur->dirty)
			continue;
		if (p_cur->dma_addr) {
			dma_sync_single_for_device(dev, p_cur->dma_addr,
						   E50_PAGE_SIZE,
						   DMA_TO_DEVICE);
			p_cur->dirty = 0;
		}
	}
}

static int mmu_map_buffer(struct device *dev, struct list_head *pentry_list,
			  struct page_entry *p_top, int hugepages,
			  uint64_t vaddr_mmu_dsp, struct elcore50_buf_desc *buf)
{
	/*
	 * the only userptr that specifies the vaddr is ELF, and addresses
	 * for it are reserved in the dsp_pool.
	 */
	/* TODO: sync the memory? create non-cacheable memory? */
	int rc = 0;
	enum pte_type type;
	uint64_t vaddr_start_mmu_dsp;
	int i;
	struct scatterlist *iter;
	struct userptr_mapper *mapper = buf->mapper;
	dma_addr_t dmastart;
	size_t dmalen, dma_offset;

	if (buf->wr_addr)
		*buf->wr_addr = vaddr_mmu_dsp + mapper->offset;

	switch (buf->type) {
	case ELCORE50_ELF_SECTION_CODE:
		type = PTE_SVRWXURWX;
		break;
	case ELCORE50_ELF_SECTION_DATA:
		type = PTE_SVRWXURWX;
		break;
	case ELCORE50_ELF_SECTION_DATA_CONST:
		type = PTE_SVRWXURWX;
	}
	vaddr_start_mmu_dsp = vaddr_mmu_dsp;
	for_each_sg(mapper->sgt->sgl, iter, mapper->sgt->nents, i) {
		dma_offset = sg_dma_address(iter) & E50_INPAGE_MASK;
		dmastart = sg_dma_address(iter) - dma_offset;
		dmalen = round_up(sg_dma_len(iter) + dma_offset, E50_PAGE_SIZE);
		pr_debug("dmastart=%llx len=%lx ofst=%lx\n", dmastart, dmalen,
			 dma_offset);
		rc = write_buffer_pages(dev, hugepages, pentry_list, p_top,
					vaddr_mmu_dsp, type, dmastart, dmalen,
					buf->arg_type ==
						ELCORE50_TYPE_DMA_MEMORY,
					mapper, vaddr_start_mmu_dsp);
		if (rc)
			return rc;
		vaddr_mmu_dsp += dmalen;
	}
	return rc;
}

static int mmu_map_elf(struct elcore50_job_desc *job_desc,
		       struct elcore50_buf_desc *elf)
{
	uint64_t vaddr_mmu_dsp;

	vaddr_mmu_dsp = elcore50_mmu_alloc_fixed(job_desc->dsp_pool,
						 elf->mapper->size_aligned,
						 elf->vaddr_mmu_dsp);
	if (!vaddr_mmu_dsp)
		return -ENOMEM;
	elf->vaddr_mmu_dsp = vaddr_mmu_dsp;
	elf->mmu_allocated = 1;
	elf->mapper->is_argument = 0;
	trace_mmu_map(elf);
	return mmu_map_buffer(JOB_DESC_TO_DEV(job_desc),
			      &job_desc->page_entry_list, job_desc->p_top,
			      job_desc->hugepages, vaddr_mmu_dsp, elf);
}

static int mmu_map_buf(struct elcore50_job_inst_desc *job_inst,
		       struct elcore50_buf_desc *buf)
{
	struct elcore50_job_desc *job_desc = job_inst->job_desc;
	uint64_t vaddr_mmu_dsp;
	struct mmu_pool *pool;
	struct device *dev = JOB_DESC_TO_DEV(job_desc);

	if (buf->arg_type == ELCORE50_TYPE_DMA_MEMORY)
		pool = job_inst->dma_pool;
	else
		pool = job_inst->dsp_pool;
	if (buf->arg_type == ELCORE50_TYPE_NC_GLOBAL_MEMORY) {
		vaddr_mmu_dsp = elcore50_mmu_alloc_fixed(
			pool, buf->mapper->size_aligned,
			job_inst->nc_mem_current);
		job_inst->nc_mem_current += buf->mapper->size_aligned;
	} else
		vaddr_mmu_dsp = mmu_pool_alloc(pool, buf->mapper->size_aligned);
	buf->vaddr_mmu_dsp = vaddr_mmu_dsp;
	if (!vaddr_mmu_dsp)
		return -ENOMEM;
	buf->mmu_allocated = 1;
	buf->mapper->is_argument = 1;
	trace_mmu_map(buf);
	return mmu_map_buffer(dev, &job_desc->page_entry_list, job_desc->p_top,
			      job_desc->hugepages, vaddr_mmu_dsp, buf);
}

static void elcore50_mmu_free_fixed(struct elcore50_job_desc *job_desc)
{
	int cluster_idx, i, size;
	unsigned long addr;

	size = sizeof(protected_regions) / sizeof(struct e50_protected_region);

	for (i = size - 1; i >= 0; i--) {
		for (cluster_idx = 0;
		     cluster_idx < elcore50_platform_info.cluster_cap;
		     cluster_idx++) {
			addr = protected_regions[i].base_addr;
			if (protected_regions[i].is_cluster_specific)
				addr += cluster_idx * protected_regions[i].size;
			else if (cluster_idx)
				break;

			mmu_pool_free(job_desc->dsp_pool, addr,
				      protected_regions[i].size);
		}
	}
}

static int elcore50_mmu_fill_fixed(struct elcore50_job_desc *job_desc)
{
	int cluster_idx, i, size, ret;
	unsigned long addr;
	dma_addr_t dmastart;
	struct device *dev = JOB_DESC_TO_DEV(job_desc);
	struct userptr_mapper *mapper;
	size_t dmalen, dma_offset;

	size = sizeof(protected_regions) / sizeof(struct e50_protected_region);

	for (i = 0; i < size; ++i) {
		for (cluster_idx = 0;
		     cluster_idx < elcore50_platform_info.cluster_cap;
		     cluster_idx++) {
			addr = protected_regions[i].base_addr;
			if (protected_regions[i].is_cluster_specific)
				addr += cluster_idx * protected_regions[i].size;
			else if (cluster_idx)
				break;

			addr = elcore50_mmu_alloc_fixed(
				job_desc->dsp_pool, protected_regions[i].size,
				addr);
			if (!addr)
				goto fixed_alloc_err;

			dma_offset = addr & E50_INPAGE_MASK;
			dmastart = addr - dma_offset;
			dmalen =
				round_up(protected_regions[i].size + dma_offset,
					 E50_PAGE_SIZE);
			mapper = &protected_regions[i].mapper;
			ret = write_buffer_pages(dev, job_desc->hugepages,
						 &job_desc->page_entry_list,
						 job_desc->p_top, addr,
						 PTE_SVRWXURWX, dmastart,
						 dmalen, 0, mapper, addr);
			if (ret) {
				mmu_pool_free(job_desc->dsp_pool, addr,
					      protected_regions[i].size);
				goto fixed_alloc_err;
			}
		}
	}
	return 0;

fixed_alloc_err:
	while (i) {
		while (cluster_idx) {
			addr = protected_regions[i - 1].base_addr;
			if (protected_regions[i - 1].is_cluster_specific)
				addr += cluster_idx *
					protected_regions[i - 1].size;
			else if (cluster_idx) {
				cluster_idx--;
				continue;
			}

			mmu_pool_free(job_desc->dsp_pool, addr,
				      protected_regions[i - 1].size);
			cluster_idx--;
		}
		i--;
		cluster_idx = elcore50_platform_info.cluster_cap - 1;
	}
	return -ENOMEM;
}

static int elcore50_is_addr_xyram(struct elcore50_job_desc *job_desc,
				  struct elcore50_buf_desc *elf)
{
	unsigned long start, end;

	start = elf->vaddr_mmu_dsp;
	end = start + elf->mapper->size;

	if ((start >= PHYS_INTERNAL_DSP) &&
	    (end <= (PHYS_INTERNAL_INTERLEAVE_DSP + INTERNAL_DSP_SIZE))) {
		return 1;
	}
	return 0;
}

static void elcore50_prepare_xyram(struct elcore50_job_desc *job_desc,
				   struct elcore50_buf_desc *elf)
{
	struct userptr_mapper *mapper = elf->mapper;

	if (!mapper->vaddr)
		mapper->vaddr = dma_buf_vmap_compat(mapper->attach->dmabuf);

	job_desc->xyram = kzalloc(sizeof(struct elcore50_buf_desc), GFP_KERNEL);
	job_desc->xyram->mapper = elf->mapper;
	job_desc->xyram->vaddr_mmu_dsp = elf->vaddr_mmu_dsp;
}

static int elcore50_is_addr_pram(struct elcore50_job_desc *job_desc,
				 struct elcore50_buf_desc *elf)
{
	unsigned long start, end;

	start = elf->vaddr_mmu_dsp;
	end = start + elf->mapper->size;

	if ((start >= PHYS_INTERNAL_PRAM_DSP) &&
	    (end <= (PHYS_INTERNAL_PRAM_DSP + INTERNAL_DSP_PRAM_SIZE))) {
		return 1;
	}
	return 0;
}

static void elcore50_prepare_pram(struct elcore50_job_desc *job_desc,
				  struct elcore50_buf_desc *elf)
{
	struct userptr_mapper *mapper = elf->mapper;

	if (!mapper->vaddr)
		mapper->vaddr = dma_buf_vmap_compat(mapper->attach->dmabuf);

	job_desc->pram = kzalloc(sizeof(struct elcore50_buf_desc), GFP_KERNEL);
	job_desc->pram->mapper = elf->mapper;
	job_desc->pram->vaddr_mmu_dsp = elf->vaddr_mmu_dsp;
}

void elcore50_mmu_free(struct elcore50_job_desc *job_desc)
{
	int i;
	struct elcore50_buf_desc *elf;
	struct device *dev = JOB_DESC_TO_DEV(job_desc);

	__free_page(job_desc->reserve_page);

	elcore50_mmu_free_fixed(job_desc);

	dma_unmap_single_attrs(dev, job_desc->pt4_dma_addr, E50_PAGE_SIZE,
			       DMA_TO_DEVICE, DMA_ATTR_SKIP_CPU_SYNC);

	mmu_free_page_entries(dev, &job_desc->page_entry_list, 0);
	free_pages((unsigned long)job_desc->pt4, get_order(E50_PAGE_SIZE));

	for (i = 0; i < job_desc->num_elf_sections; i++) {
		elf = job_desc->elf[i];
		if (elcore50_is_addr_xyram(job_desc, elf))
			continue;
		if (elcore50_is_addr_pram(job_desc, elf))
			continue;
		mmu_pool_free(job_desc->dsp_pool, elf->vaddr_mmu_dsp,
			      elf->mapper->size_aligned);
	}
}

void elcore50_mmu_free_args(struct elcore50_job_inst_desc *job_inst)
{
	int i;
	struct elcore50_buf_desc *buf;
	struct phys_mapper *cursor, *store;

	list_for_each_entry_safe(cursor, store, &job_inst->phys_mappers,
				 queue_node) {
		elcore50_release_physmapper(job_inst, cursor);
	}

	mmu_free_page_entries(job_inst->job_desc->dev,
			      &job_inst->job_desc->page_entry_list, 1);

	for (i = 0; i < job_inst->argc; i++) {
		buf = job_inst->args[i];
		if (buf->arg_type == ELCORE50_TYPE_DMA_MEMORY)
			mmu_pool_free(job_inst->dma_pool, buf->vaddr_mmu_dsp,
				      buf->mapper->size_aligned);
		else if (buf->arg_type == ELCORE50_TYPE_NC_GLOBAL_MEMORY)
			mmu_pool_free(job_inst->dsp_pool, buf->vaddr_mmu_dsp,
				      buf->mapper->size_aligned);
		else
			mmu_pool_free(job_inst->dsp_pool, buf->vaddr_mmu_dsp,
				      buf->mapper->size_aligned);
	}
}

static void elcore50_add_prefetch_pages(struct elcore50_job_inst_desc *job_inst,
					u64 vaddr_mmu_dsp, size_t size)
{
	uint32_t addr;
	bool found;
	dma_addr_t paddr = page_to_phys(job_inst->job_desc->reserve_page);

	if (vaddr_mmu_dsp & (SZ_2M - 1)) {
		found = addr_in_mmu_pool_chunk(job_inst->dsp_pool,
					       vaddr_mmu_dsp - SZ_4K, SZ_4K);
		if (!found) {
			// Map 4K-page at (vaddr_mmu_dsp - SZ_4K) for
			// backward cache prefetch
			addr = elcore50_syscall_pmmap(job_inst,
						      vaddr_mmu_dsp - SZ_4K,
						      SZ_4K, paddr, true);
			WARN_ON(addr != (vaddr_mmu_dsp - SZ_4K));
		}
	}

	if ((vaddr_mmu_dsp + size) & (SZ_2M - 1)) {
		found = addr_in_mmu_pool_chunk(job_inst->dsp_pool,
					       vaddr_mmu_dsp + size, SZ_4K);

		if (!found) {
			// Map 4K-page at (vaddr_mmu_dsp + size) for
			// forward cache prefetch
			addr = elcore50_syscall_pmmap(job_inst,
						      vaddr_mmu_dsp + size,
						      SZ_4K, paddr, true);
			WARN_ON(addr != (vaddr_mmu_dsp + size));
		}
	}
}

void elcore50_mmu_sync(struct elcore50_job_inst_desc *job_inst)
{
	int i;
	struct elcore50_job_desc *job_desc = job_inst->job_desc;
	struct device *dev = JOB_DESC_TO_DEV(job_desc);
	size_t size;
	u64 vaddr_mmu_dsp;

	// Add empty 4K pages to prevent cache prefetch error
	for (i = 0; i < job_desc->num_elf_sections; ++i) {
		vaddr_mmu_dsp = job_desc->elf[i]->vaddr_mmu_dsp;
		size = job_desc->elf[i]->mapper->size_aligned;

		elcore50_add_prefetch_pages(job_inst, vaddr_mmu_dsp, size);
	}

	for (i = 0; i < job_inst->argc; ++i) {
		if (job_inst->args[i]->arg_type == ELCORE50_TYPE_DMA_MEMORY)
			continue;
		vaddr_mmu_dsp = job_inst->args[i]->vaddr_mmu_dsp;
		size = job_inst->args[i]->mapper->size_aligned;

		elcore50_add_prefetch_pages(job_inst, vaddr_mmu_dsp, size);
	}

	dma_sync_single_for_device(dev, job_inst->job_desc->pt4_dma_addr,
				   E50_PAGE_SIZE, DMA_TO_DEVICE);
	sync_elcore_pte_table_for_device(dev, &job_desc->page_entry_list);
}

int elcore50_mmu_fill(struct elcore50_job_desc *job_desc)
{
	int i, rc = 0;
	uint64_t *pt4;
	struct device *dev = JOB_DESC_TO_DEV(job_desc);
	struct page_entry *p_top;
	struct page *page;

	page = alloc_pages(GFP_KERNEL | __GFP_ZERO, get_order(E50_PAGE_SIZE));
	if (!page)
		return -ENOMEM;
	pt4 = page_address(page);
	job_desc->pt4 = pt4;
	job_desc->pt4_dma_addr = dma_map_single_attrs(
		dev, pt4, E50_PAGE_SIZE, DMA_TO_DEVICE, DMA_ATTR_SKIP_CPU_SYNC);
	if (!job_desc->pt4_dma_addr) {
		free_pages((unsigned long)pt4, get_order(E50_PAGE_SIZE));
		return -ENOMEM;
	}

	p_top = kcalloc(E50_PTE_ENTRIES, sizeof(struct page_entry), GFP_KERNEL);
	if (!p_top) {
		rc = -ENOMEM;
		goto clean_pt4;
	}
	/* Set to prevent p_top page releasing in mmu_free_page_entries(). */
	p_top->num_non_zero_ptes = 1;
	job_desc->p_top = p_top;
	for (i = 0; i < E50_PTE_ENTRIES; i++)
		p_top[i].pte = &pt4[i];

	rc = elcore50_mmu_fill_fixed(job_desc);
	if (rc) {
		kfree(p_top);
		goto clean_pt4;
	}

	for (i = 0; i < job_desc->num_elf_sections; i++) {
		if (elcore50_is_addr_xyram(job_desc, job_desc->elf[i])) {
			elcore50_prepare_xyram(job_desc, job_desc->elf[i]);
			continue;
		}
		if (elcore50_is_addr_pram(job_desc, job_desc->elf[i])) {
			elcore50_prepare_pram(job_desc, job_desc->elf[i]);
			continue;
		}
		rc = mmu_map_elf(job_desc, job_desc->elf[i]);
		if (rc != 0)
			goto clean_elf;
	}

	job_desc->reserve_page =
		alloc_pages(GFP_KERNEL, get_order(E50_PAGE_SIZE));
	if (!job_desc->reserve_page)
		goto clean_elf;

	/* Here we can to sync ptes for ELF sections */
	sync_elcore_pte_table_for_device(dev, &job_desc->page_entry_list);

	return 0;
clean_elf:
	for (i = 0; i < job_desc->num_elf_sections; i++) {
		if (job_desc->elf[i]->mmu_allocated)
			mmu_pool_free(job_desc->dsp_pool,
				      job_desc->elf[i]->vaddr_mmu_dsp,
				      job_desc->elf[i]->mapper->size_aligned);
	}
	elcore50_mmu_free_fixed(job_desc);
clean_pt4:
	mmu_free_page_entries(dev, &job_desc->page_entry_list, 0);
	dma_unmap_single_attrs(dev, (unsigned long)pt4, E50_PAGE_SIZE,
			       DMA_TO_DEVICE, DMA_ATTR_SKIP_CPU_SYNC);
	free_pages((unsigned long)pt4, get_order(E50_PAGE_SIZE));
	return rc;
}

int elcore50_mmu_fill_args(struct elcore50_job_inst_desc *job_inst)
{
	int i, rc = 0;
	struct elcore50_buf_desc *arg;
	struct userptr_mapper *mapper;
	uint64_t vaddr_mmu_dsp = 0;

	for (i = 0; i < job_inst->argc; i++) {
		arg = job_inst->args[i];
		if (arg->arg_type != ELCORE50_TYPE_NC_GLOBAL_MEMORY)
			continue;
		rc = mmu_map_buf(job_inst, arg);
		if (rc != 0)
			goto clean_args;
	}
	if (job_inst->nc_mem_current) {
		vaddr_mmu_dsp = elcore50_mmu_alloc_fixed(
			job_inst->dsp_pool,
			(1UL << 32) - job_inst->nc_mem_current,
			job_inst->nc_mem_current);
		if (!vaddr_mmu_dsp) {
			rc = -ENOMEM;
			goto clean_args;
		}
	}

	for (i = 0; i < job_inst->argc; i++) {
		arg = job_inst->args[i];
		if (arg->arg_type == ELCORE50_TYPE_NC_GLOBAL_MEMORY)
			continue;
		rc = mmu_map_buf(job_inst, arg);
		if (rc != 0)
			goto clean_args;
	}

	if (job_inst->nc_mem_current && vaddr_mmu_dsp) {
		mmu_pool_free(job_inst->dsp_pool, vaddr_mmu_dsp,
			      (1UL << 32) - job_inst->nc_mem_current);
	}
	return 0;
clean_args:
	for (i = 0; i < job_inst->argc; i++) {
		if (job_inst->args[i]->mmu_allocated) {
			arg = job_inst->args[i];
			mapper = arg->mapper;
			if (arg->arg_type == ELCORE50_TYPE_DMA_MEMORY)
				mmu_pool_free(job_inst->dma_pool,
					      arg->vaddr_mmu_dsp,
					      mapper->size_aligned);
			else
				mmu_pool_free(job_inst->dsp_pool,
					      arg->vaddr_mmu_dsp,
					      mapper->size_aligned);
		}
	}
	if (job_inst->nc_mem_current && vaddr_mmu_dsp) {
		mmu_pool_free(job_inst->dma_pool, vaddr_mmu_dsp,
			      (1UL << 32) - job_inst->nc_mem_current);
	}
	return rc;
}

static bool check_protected_regions_overlapping(u64 paddr, size_t length)
{
	size_t cluster_idx, i, protected_regions_num;
	u64 region_start_addr, region_end_addr;

	protected_regions_num =
		sizeof(protected_regions) / sizeof(struct e50_protected_region);

	for (i = 0; i < protected_regions_num; ++i) {
		for (cluster_idx = 0;
		     cluster_idx < elcore50_platform_info.cluster_cap;
		     cluster_idx++) {
			region_start_addr = (u64)protected_regions[i].base_addr;
			if (protected_regions[i].is_cluster_specific)
				region_start_addr +=
					cluster_idx * protected_regions[i].size;
			else if (cluster_idx)
				break;
			region_end_addr =
				region_start_addr + protected_regions[i].size;

			if ((region_start_addr <= (paddr + length)) &&
			    (paddr <= region_end_addr)) {
				/* Physical range overlaps protected region
				 * that can only be accessed by a physical
				 * address.
				 */
				return true;
			}
		}
	}
	return false;
}

uint32_t elcore50_syscall_pmmap(struct elcore50_job_inst_desc *job_inst,
				uint32_t pref_vaddr, size_t length,
				uint64_t paddr, bool is_reserved)
{
	struct elcore50_job_desc *job_desc = job_inst->job_desc;
	struct device *dev = JOB_DESC_TO_DEV(job_desc);
	struct phys_mapper *pmapper;
	uint64_t vaddr_mmu_dsp = 0;
	dma_addr_t dmastart;
	size_t dmalen, dma_offset;
	int ret;

	if (length & E50_INPAGE_MASK)
		return 0;

	ret = check_protected_regions_overlapping(paddr, length);
	if (ret)
		return 0;

	if (pref_vaddr)
		vaddr_mmu_dsp = elcore50_mmu_alloc_fixed(job_desc->dsp_pool,
							 length, pref_vaddr);
	if (!vaddr_mmu_dsp)
		vaddr_mmu_dsp = mmu_pool_alloc(job_desc->dsp_pool, length);

	if (!vaddr_mmu_dsp) {
		// TODO: Try to free space by removing prefetch reserved
		//       regions.
		return 0;
	}

	pmapper = kmalloc(sizeof(struct phys_mapper), GFP_KERNEL);
	if (!pmapper)
		goto pool_free_vaddr;

	pmapper->paddr = paddr;
	pmapper->size = length;
	pmapper->vaddr = vaddr_mmu_dsp;
	pmapper->mapper_stub.is_argument = 1;
	pmapper->is_reserved = is_reserved;

	if (!is_reserved)
		elcore50_add_prefetch_pages(job_inst, vaddr_mmu_dsp, length);

	dma_offset = paddr & E50_INPAGE_MASK;
	dmastart = paddr - dma_offset;
	dmalen = round_up(length + dma_offset, E50_PAGE_SIZE);
	ret = write_buffer_pages(dev, job_desc->hugepages,
				 &job_desc->page_entry_list, job_desc->p_top,
				 vaddr_mmu_dsp, PTE_SVRWXURWX, dmastart, dmalen,
				 0, &pmapper->mapper_stub, vaddr_mmu_dsp);
	if (ret)
		goto pmapper_kfree;

	list_add_tail(&pmapper->queue_node, &job_inst->phys_mappers);

	return vaddr_mmu_dsp;
pmapper_kfree:
	kfree(pmapper);
pool_free_vaddr:
	mmu_pool_free(job_desc->dsp_pool, vaddr_mmu_dsp, length);
	return 0;
}

void elcore50_release_physmapper(struct elcore50_job_inst_desc *job_inst,
				 struct phys_mapper *cursor)
{
	mmu_pool_free(job_inst->dsp_pool, cursor->vaddr, cursor->size);

	/* page entries will be released in mmu_free_page_entries() */

	list_del(&cursor->queue_node);
	kfree(cursor);
}

int elcore50_syscall_pmunmap(struct elcore50_job_inst_desc *job_inst,
			     uint32_t vaddr_mmu_dsp, size_t length)
{
	struct phys_mapper *cursor, *store;
	uint32_t cur_vaddr;
	size_t cur_size;
	bool found;

	list_for_each_entry_safe(cursor, store, &job_inst->phys_mappers,
				 queue_node) {
		cur_vaddr = cursor->vaddr;
		cur_size = cursor->size;
		if ((cur_vaddr == vaddr_mmu_dsp) && (cur_size == length)) {
			mmu_pool_free(job_inst->dsp_pool, vaddr_mmu_dsp,
				      length);
			list_del(&cursor->queue_node);
			kfree(cursor);
			goto goto_found;
		}
	}
	return -EINVAL;
goto_found:
	// Remove prefetch protected regions
	list_for_each_entry_safe(cursor, store, &job_inst->phys_mappers,
				 queue_node) {
		if (!cursor->is_reserved)
			continue;
		cur_vaddr = cursor->vaddr;
		cur_size = cursor->size;

		WARN_ON(cur_size != SZ_4K);

		if (cur_vaddr == (vaddr_mmu_dsp - SZ_4K)) {
			found = addr_in_mmu_pool_chunk(
				job_inst->dsp_pool, cur_vaddr - SZ_4K, SZ_4K);
			if (!found)
				elcore50_release_physmapper(job_inst, cursor);
		}
		if (cur_vaddr == (vaddr_mmu_dsp + length)) {
			found = addr_in_mmu_pool_chunk(
				job_inst->dsp_pool, cur_vaddr + SZ_4K, SZ_4K);
			if (!found)
				elcore50_release_physmapper(job_inst, cursor);
		}
	}
	return 0;
}

inline bool elcore50_vaddr_is_internal(u64 vaddr_mmu_dsp, size_t size)
{
	const uint32_t internal_end_addr =
		PHYS_INTERNAL_INTERLEAVE_DSP + INTERNAL_DSP_SIZE;

	if ((vaddr_mmu_dsp >= PHYS_INTERNAL_DSP) &&
	    (vaddr_mmu_dsp + size) <= internal_end_addr)
		return true;
	return false;
}

static struct userptr_mapper *
elcore50_vaddr_to_mapper_internal(struct page_entry *p, u64 vaddr_mmu_dsp,
				  u8 *out_inpage_offset, size_t *out_pte_offset)
{
	u64 index, i;
	u8 inpage_offset = 38;

	if (p == NULL) {
		WARN_ON(1);
		return NULL;
	}

	for (i = 3; i >= 0; i--) {
		index = (vaddr_mmu_dsp >> (E50_PAGE_SHIFT + i * 9)) &
			GENMASK(8, 0);
		p = p + index;
		if (p->next_lvl == NULL)
			break;
		p = p->next_lvl;
		inpage_offset -= 9;
	}

	if ((p == NULL) || (p->mapper == NULL)) {
		WARN_ON(1);
		return NULL;
	}

	if (out_inpage_offset)
		*out_inpage_offset = inpage_offset;

	if (out_pte_offset)
		*out_pte_offset = p->offset;

	return p->mapper;
}

struct userptr_mapper *elcore50_vaddr_to_mapper(struct elcore50_job_desc *desc,
						u64 vaddr_mmu_dsp)
{
	struct userptr_mapper *mapper;

	if (elcore50_vaddr_is_internal(vaddr_mmu_dsp, 0))
		mapper = desc->xyram->mapper;
	else
		mapper = elcore50_vaddr_to_mapper_internal(
			desc->p_top, vaddr_mmu_dsp, NULL, NULL);

	if (mapper && !mapper->vaddr)
		mapper->vaddr = dma_buf_vmap_compat(mapper->attach->dmabuf);

	return mapper;
}

void *elcore50_vaddr_to_kernel(struct elcore50_job_desc *desc,
			       u64 vaddr_mmu_dsp)
{
	bool is_internal = elcore50_vaddr_is_internal(vaddr_mmu_dsp, 0);
	void *retval;
	u8 inpage_offset;
	size_t pte_offset = 0;
	struct userptr_mapper *mapper;

	if (is_internal)
		mapper = desc->xyram->mapper;
	else
		mapper = elcore50_vaddr_to_mapper_internal(desc->p_top,
							   vaddr_mmu_dsp,
							   &inpage_offset,
							   &pte_offset);
	if (!mapper)
		return NULL;

	if (!mapper->vaddr)
		mapper->vaddr = dma_buf_vmap_compat(mapper->attach->dmabuf);

	retval = mapper->vaddr + pte_offset - mapper->offset;
	if (is_internal)
		retval += vaddr_mmu_dsp - PHYS_INTERNAL_DSP;
	else
		retval += vaddr_mmu_dsp & GENMASK(inpage_offset, 0);

	return retval;
}

/* Translate DSP virtual address to CPU virtual address */
void *elcore50_map_from_users(struct elcore50_job_inst_desc *desc,
			      u64 vaddr_mmu_dsp,
			      struct userptr_mapper **out_mapper,
			      size_t *offset, u64 *user_vaddr_cpu, size_t size)
{
	bool is_internal = elcore50_vaddr_is_internal(vaddr_mmu_dsp, size);
	u64 ret;
	void *retval;
	u8 inpage_offset;
	size_t pte_offset = 0;
	struct userptr_mapper *mapper;
	struct elcore50_buffer_priv *buf_priv;
	struct elcore50_core *core = desc->core;
	struct elcore50_job_desc *job_desc = desc->job_desc;

	WARN_ON(size == 0);

	if (is_internal)
		mapper = job_desc->xyram->mapper;
	else
		mapper = elcore50_vaddr_to_mapper_internal(job_desc->p_top,
							   vaddr_mmu_dsp,
							   &inpage_offset,
							   &pte_offset);
	if (!mapper)
		return NULL;

	if (mapper->dmabuf->ops != &elcore50_dmabuf_ops) {
		dev_err(core->drv_priv->dev,
			"Attempt to map external buffer\n");
		return NULL;
	}
	buf_priv = mapper->dmabuf->priv;

	*out_mapper = mapper;
	if (!mapper->vaddr)
		mapper->vaddr = dma_buf_vmap_compat(mapper->attach->dmabuf);
	retval = mapper->vaddr + pte_offset - mapper->offset;
	if (is_internal)
		retval += vaddr_mmu_dsp - PHYS_INTERNAL_DSP;
	else
		retval += vaddr_mmu_dsp & GENMASK(inpage_offset, 0);
	*offset = retval - mapper->vaddr;
	if (buf_priv->buf_info.p == 0)
		*user_vaddr_cpu = 0;
	else
		*user_vaddr_cpu = buf_priv->buf_info.p + *offset;

	if (is_internal && (vaddr_mmu_dsp >= PHYS_INTERNAL_INTERLEAVE_DSP)) {
		retval -= INTERNAL_DSP_INTERLEAVE;

		if (*user_vaddr_cpu)
			*user_vaddr_cpu -= INTERNAL_DSP_INTERLEAVE;
	}

	if (is_internal)
		memcpy_fromio(retval, core->xyram + *offset, size);
	else {
		// Invalidate CPU caches
		ret = sync_buffer(size, *offset, mapper,
				  ELCORE50_BUF_SYNC_DIR_TO_CPU);
		if (ret)
			return NULL;
	}

	return retval;
}
