// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright 2024 RnD Center "ELVEES", JSC
 */

#include <linux/seq_file.h>

#include "elcore50-pm.h"
#include "elcore50-proc.h"

static int elcore50_proc_show_usage(struct seq_file *m, void *v)
{
	int i = 0, int_usage, fract_usage;
	struct elcore50_core **cores = &scheduler_data.core[0];

	while (cores[i]) {
		/* In Solaris system timer is unstable (SOLARISSW-1222).
		 * So usage > 100% is possible. For timer period = 500 msec
		 * the fluctuation is insignificant.
		 */
		if (cores[i]->current_usage > 10000) {
			int_usage = 100;
			fract_usage = 0;
		} else {
			int_usage = cores[i]->current_usage / 100;
			fract_usage = cores[i]->current_usage % 100;
		}
		seq_printf(m, "%u.%02u ", int_usage, fract_usage);
		i++;
	}
	seq_puts(m, "\n");
	return 0;
}

static int elcore50_proc_show_info(struct seq_file *m, void *v)
{
	int i = 0;
	struct elcore50_core **cores = &scheduler_data.core[0];
	u32 idr_val;

	while (cores[i]) {
		if (i)
			seq_puts(m, "\n");

		idr_val = cores[i]->idr;

		seq_printf(m, "%-19s: %d\n", "processor", cores[i]->system_idx);
		seq_printf(m, "%-19s: ELcore-50\n", "DSP family");
		seq_printf(m, "%-19s: 0x%02lx\n", "DSP revision",
			   IDR_VER(idr_val));
		seq_printf(m, "%-19s: 0x%02lx\n", "DSP subrevision",
			   IDR_REV(idr_val));
		seq_printf(m, "%-19s: %ld\n", "cluster", IDR_CLUSTER(idr_val));
		seq_printf(m, "%-19s: %ld\n", "position in cluster",
			   IDR_CORENUM(idr_val));
		i++;
	}
	return 0;
}

static int elcore50_proc_usage_open(struct inode *inode, struct file *file)
{
	return single_open_size(file, elcore50_proc_show_usage,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 17, 0)
				pde_data(file_inode(file)), PAGE_SIZE);
#else
				PDE_DATA(file_inode(file)), PAGE_SIZE);
#endif
}

static int elcore50_proc_info_open(struct inode *inode, struct file *file)
{
	return single_open_size(file, elcore50_proc_show_info,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 17, 0)
				pde_data(file_inode(file)), PAGE_SIZE);
#else
				PDE_DATA(file_inode(file)), PAGE_SIZE);
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 6, 0)
static const struct file_operations proc_usage_fops = {
	.owner = THIS_MODULE,
	.open = elcore50_proc_usage_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};

static const struct file_operations proc_info_fops = {
	.owner = THIS_MODULE,
	.open = elcore50_proc_info_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};
#else
static const struct proc_ops proc_usage_fops = {
	.proc_open = elcore50_proc_usage_open,
	.proc_read = seq_read,
	.proc_lseek = seq_lseek,
	.proc_release = single_release,
};

static const struct proc_ops proc_info_fops = {
	.proc_open = elcore50_proc_info_open,
	.proc_read = seq_read,
	.proc_lseek = seq_lseek,
	.proc_release = single_release,
};
#endif

int elcore50_proc_init(void)
{
	struct proc_dir_entry *pde;

	pde = proc_create_data("usage", 0444, proc_pdentry, &proc_usage_fops,
			       NULL);
	if (!pde)
		return -ENOMEM;

	pde = proc_create_data("info", 0444, proc_pdentry, &proc_info_fops,
			       NULL);
	if (!pde)
		return -ENOMEM;

	return 0;
}
